#include "stdafx.h"

HANDLE wHnd;
HANDLE rHnd;

SMALL_RECT windowSize = {0, 0, 79, 49};
COORD bufferSize = {80, 50};
COORD charBufSize = {80, 50};
COORD characterPos = {0,0};
SMALL_RECT writeArea = {0,0,79,49}; 

CHAR_INFO buffers[3][80*50];
CHAR_INFO *console_buffer;

CHAR_INFO block_palette[8*8];

BOOL key_pressed[256] = {false,};

CHAR_INFO *palette;

char buffer_index = 0;

int *rgb_map;
short *colour_map;

short quick_rgb_map[0xFFF] = {0,};

short num_colours;

short find_closest_colour(int r, int g, int b) {
	int closest_d = 99999999;
	int closest_n = -1;
	for (int i=0; i<num_colours; ++i) {
		int c = rgb_map[i];
		int c_b = c&0xFF;
		c>>=8;
		int c_g = c&0xFF;
		c>>=8;
		int c_r = c&0xFF;
		int dist = abs(c_r-r)+abs(c_g-g)+abs(c_b-b);
		if (dist<closest_d) {
			closest_d = dist;
			closest_n = i;
		}
	}
	return closest_n;
}

short quick_find_closest_colour(int r, int g, int b) {
	if (r>0xFF) r = 0xFF;
	if (g>0xFF) g = 0xFF;
	if (b>0xFF) b = 0xFF;
	if (r<0x00) r = 0x00;
	if (g<0x00) g = 0x00;
	if (b<0x00) b = 0x00;	
	return quick_rgb_map[(b>>4)+(g&0xF0)+((r&0xF0)<<4)];
}

bool save_screenshot(char *filename) {
	
	FILE *rtf = fopen(filename, "wb");
	if (!rtf) {
		return false;
	}
	fprintf(rtf, "{\\rtf1\\ansi\\ansicpg1252\\deff0\\deflang2057{\\fonttbl{\\f0\\fmodern\\fprq1\\fcharset0 Courier New;}}\n{\\colortbl ");
    
	for (int i=0; i<128; i+=127) for (int r=0; r<128; r+=127) for (int g=0; g<128; g+=127) for (int b=0; b<128; b+=127) {
		fprintf(rtf, "\\red%d\\green%d\\blue%d;", r + i, g + i, b + i);
	}
    fprintf(rtf, "}\n{\\*\\generator ASCIIMAD2;}\\viewkind4\\uc1\\pard\\f0\\fs12 ");
	
    char oBackC = 17;
    char oForeC = 17;
    char nBackC = 17;
    char nForeC = 17;
	
	CHAR_INFO temp_buffer[80*50];
	ReadConsoleOutput(wHnd,temp_buffer,charBufSize,characterPos,&writeArea);	

	for (int y=0; y<50; ++y) {
		for (int x=0; x<80; ++x) {
			int colour = temp_buffer[x+y*80].Attributes;
            nForeC = colour&0xF;
            nBackC = (colour>>4)&0xF;

			if (nForeC!=oForeC) {
				oForeC = nForeC;
				fprintf(rtf, "\\cf%d ", nForeC);
			}
			if (nBackC!=oBackC) {
				oBackC = nBackC;
				fprintf(rtf, "\\highlight%d ", nBackC);
			}
			int asc_char = temp_buffer[x+y*80].Char.AsciiChar;
			asc_char &= 0xFF;
            fprintf(rtf, "\\u%d?", 9441+asc_char);
		}
        if (y!=49) fprintf(rtf,"\\par ");
	}
	fprintf(rtf,"}");
	fclose(rtf);
	return true;
}


void hide_cursor() {
	CONSOLE_CURSOR_INFO h_cursor;
	h_cursor.bVisible = false;
	h_cursor.dwSize = 1;
	SetConsoleCursorInfo(wHnd, &h_cursor);
}

bool set_up_console() {
    wHnd = GetStdHandle(STD_OUTPUT_HANDLE);
    rHnd = GetStdHandle(STD_INPUT_HANDLE);
	SetConsoleTitle("ASCII Madness II");

    SetConsoleWindowInfo(wHnd, TRUE, &windowSize);
    SetConsoleScreenBufferSize(wHnd, bufferSize);
	buffer_index = 0;
	console_buffer = buffers[0];


	for (int c=0; c<8; ++c) for (int b=0; b<8; ++b) {
		int dithers[4] = {176,177,178,219};
		int colour = c;		
		if (b>3) colour|=FOREGROUND_INTENSITY;
		if (b==5 || b==3) colour|=((colour&7)<<4);
		if (b==0) colour = 0;
		block_palette[c*8+b].Attributes = colour;
		block_palette[c*8+b].Char.AsciiChar = dithers[b>>1];
	}

	// Open the palette:
	FILE *pal = fopen("data/default.pal","rb");
	if (!pal) {
		return false;
	} else {
		
		fread(&num_colours, sizeof(short), 1, pal);

		palette = new CHAR_INFO[num_colours];

		colour_map = new short[num_colours*256];
		rgb_map = new int[num_colours];

		char colour; char block; char r; char g; char b;

		for (int i=0; i<num_colours; ++i) {
			fread(&block, sizeof(char), 1, pal);
			fread(&colour, sizeof(char), 1, pal);
			fread(&r, sizeof(char), 1, pal);
			fread(&g, sizeof(char), 1, pal);
			fread(&b, sizeof(char), 1, pal);
			palette[i].Attributes = colour;
			palette[i].Char.AsciiChar = block;

			int fx_r = r;
			int fx_g = g;
			int fx_b = b;

			rgb_map[i] = (fx_b&0xFF) | ((fx_g&0xFF)<<8) | ((fx_r&0xFF)<<16);
		}

		
		fclose(pal);

		// Now we need to organise our colour map

		for (int i=0; i<num_colours; ++i) {

			int c = rgb_map[i];
			int c_b = c&0xFF;
			c>>=8;
			int c_g = c&0xFF;
			c>>=8;
			int c_r = c&0xFF;

			colour_map[i<<8] = i;

			for (int s=1; s<256; ++s) {
				int s_r = ((255-s)*c_r)/256;
				int s_g = ((255-s)*c_g)/256;
				int s_b = ((255-s)*c_b)/256;
				colour_map[s+(i<<8)] = find_closest_colour(s_r,s_g,s_b);
			}
		}

		for (int q_r=0; q_r<=0xF; ++q_r) for (int q_g=0; q_g<=0xF; ++q_g) for (int q_b=0; q_b<=0xF; ++q_b) {
			quick_rgb_map[(q_r<<8)+(q_g<<4)+q_b] = find_closest_colour((q_r<<4)+0x8,(q_g<<4)+0x8,(q_b<<4)+0x8);
		}

		for (int i=0; i<num_colours; ++i) {
			int c = rgb_map[i];
			int c_b = c&0xFF;
			c>>=8;
			int c_g = c&0xFF;
			c>>=8;
			int c_r = c&0xFF;
			quick_rgb_map[((c_b&0xF0)>>4)|((c_g&0xF0)>>0)|((c_r&0xF0)<<4)] = i;
		}	
	}
	return true;
}

void read_key_presses() {
	DWORD numEvents = 0;
	DWORD numEventsRead = 0;
	GetNumberOfConsoleInputEvents(rHnd, &numEvents);
	if (numEvents) {
		INPUT_RECORD *eventBuffer = new INPUT_RECORD[numEvents];
		ReadConsoleInput(rHnd, eventBuffer, numEvents, &numEventsRead);
		for (DWORD i=0; i<numEventsRead; ++i) {
			if (eventBuffer[i].EventType==KEY_EVENT) {
				// 'twas a key event...
				hide_cursor();
				int vk = eventBuffer[i].Event.KeyEvent.wVirtualKeyCode;
				if (vk<256) {
					key_pressed[vk] = eventBuffer[i].Event.KeyEvent.bKeyDown;
				}
			}
		}
		delete eventBuffer;
	}
}

void disp_current_buffer() {
	WriteConsoleOutput(wHnd, console_buffer, charBufSize, characterPos, &writeArea);
}

void flip_buffers() {
	buffer_index = 1-buffer_index;
	console_buffer = buffers[buffer_index];	
}

short get_palette_from_char_info(CHAR_INFO c) {
	for (int s=0; s<num_colours; ++s) {
		if (palette[s].Attributes==c.Attributes && palette[s].Char.AsciiChar==c.Char.AsciiChar) {
			return s;
			break;
		}
	}
	return 0;
}

void disp_blended_buffers(int mode, int amount_from_current) {

	if (amount_from_current<0) amount_from_current = 0;
	if (amount_from_current>0xFF) amount_from_current = 0xFF;

	CHAR_INFO *t = buffers[2];
	CHAR_INFO *c = buffers[buffer_index];
	CHAR_INFO *p = buffers[1-buffer_index];

	if (mode==0) {
		for (int y=0; y<50; ++y) for (int x=0; x<80; ++x) {
			if (rand()%254+amount_from_current>254) {
                *t++=*c++;
				p++;
			} else {
				*t++=*p++;
				c++;
			}
		}
	} else if (mode==1) {

		for (int i=0; i<80*50; ++i) {
			int c1 = rgb_map[get_palette_from_char_info(*c++)];
			int c2 = rgb_map[get_palette_from_char_info(*p++)];

			int adjusted_components[3];
			for (int a=0; a<3; ++a) {
				int b1 = (c1&0xFF)* amount_from_current;
				int b2 = (c2&0xFF)*(255-amount_from_current);
				c1>>=8; c2>>=8;
				adjusted_components[a] = (b1+b2)>>12;
				if (adjusted_components[a]<0) adjusted_components[a]=0;
				if (adjusted_components[a]>0xF) adjusted_components[a]=0xF;
			}
			*t++ = palette[quick_rgb_map[adjusted_components[0]+(adjusted_components[1]<<4)+(adjusted_components[2]<<8)]];
		}
	} else if (mode==2) {
		int wipe_amount = (80 * amount_from_current) / 256;
		
		for (int y=0; y<50; ++y) for (int x=0; x<80; ++x) {
			if (x>wipe_amount) {
				*t++=*p++; *c++;
			} else {
				*t++=*c++; *p++;
			}
		}
	} else if (mode==3) {
		int wipe_amount = (50 * amount_from_current) / 256;
		
		for (int y=0; y<50; ++y) for (int x=0; x<80; ++x) {
			if (y<wipe_amount) {
				*t++=*p++; *c++;
			} else {
				*t++=*c++; *p++;
			}
		}
	}

	WriteConsoleOutput(wHnd, buffers[2], charBufSize, characterPos, &writeArea);
}

void send_to_corner(int corner) {

	bool hi_q = true;

	int dst_x = 0;
	int dst_y = 0;
	if (corner==1||corner==3) dst_x = 40;
	if (corner==2||corner==3) dst_y = 25;
	CHAR_INFO *d = buffers[1-buffer_index]+dst_x+dst_y*80;
	CHAR_INFO *s = console_buffer;
	if (!hi_q) {
		for (int y=0; y<25; ++y) {
			for (int x=0; x<40; ++x) {
				*d++=*(s+=2);
			}
			d+=40;
			s+=80;
		}
	} else {
		for (int y=0; y<25; ++y) {
			for (int x=0; x<40; ++x) {
				int colours[4];
				int reds = 0;
				int greens = 0;
				int blues = 0;
				colours[0] = rgb_map[get_palette_from_char_info(*(s+0))];
				colours[1] = rgb_map[get_palette_from_char_info(*(s+1))];
				colours[2] = rgb_map[get_palette_from_char_info(*(s+80))];
				colours[3] = rgb_map[get_palette_from_char_info(*(s+81))];
				for (int i=0; i<4; ++i) {
					blues += colours[i]&0xFF;
					colours[i]>>=8;
					greens += colours[i]&0xFF;
					colours[i]>>=8;
					reds += colours[i]&0xFF;
				}
				blues>>=6;
				greens>>=6;
				reds>>=6;
				*d++=palette[quick_rgb_map[blues|(greens<<4)|(reds<<8)]];
				s+=2;
			}
			d+=40;
			s+=80;
		}
	}
}

void clear_current_buffer() {
	for (int i=0; i<80*50; ++i) {
		console_buffer[i].Attributes = 0;
		console_buffer[i].Char.AsciiChar = ' ';
	}
}

void set_console_colour(int foreground, int background) {
	SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE),foreground|(background<<4));
}
void reset_console_colour() {
	set_console_colour(7,0);
}
void console_response(bool OK) {
	if (OK) {
		set_console_colour(2,0);
		printf("OK!\n");
	} else {
		set_console_colour(12,0);
		printf("Failed\n");
	}
	reset_console_colour();
}
void console_warning() {
	set_console_colour(4,0);
	printf("Warning: ");
	reset_console_colour();
}

void shut_down_console() {
	if (palette) delete palette;
	if (colour_map) delete colour_map;
	if (rgb_map) delete rgb_map;
}